/** scafes_vs_petsc
 *  Copyright (c) 2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */

// Solving 2D Heat Equation on the unit square using finite differences and
// forward Euler method in parallel.
// Code is taken from PETSc example /petsc/src/ts/examples/tutorials/ex4.c
// and expanded to 2D.
// For more details see README!
// date:   18.12.2013
// author: martin.flehmig@tu-dresden.de

static char help[] = "Solves numerically the 2D heat equation).\n \
   Input parameters include:\n \
    -m <points>             Number of grid points in x direction [default=60] \n \
    -n <points>             Number of grid points in y direction [default=60] \n \
    -t <time_steps>         Number of time steps to compute [default=200] \n \
    -tt <total_time_steps>  Number of maximal time steps [default=5000] \n";

/*
   Concepts: TS^time-dependent linear problems
   Concepts: TS^heat equation
   Concepts: TS^diffusion equation
   Processors: n
*/

/*
   Include "petscdmda.h" so that we can use distributed arrays (DMDAs) to manage
   the parallel grid.  Include "petscts.h" so that we can use TS solvers.
   Note that this file automatically includes:
     petscsys.h       - base PETSc routines
     petscvec.h       - vectors         petscmat.h  - matrices
     petscis.h        - index sets      petscksp.h  - Krylov subspace methods
     petscviewer.h    - viewers         petscpc.h   - preconditioners
     petscksp.h       - linear solvers  petscsnes.h - nonlinear solvers
*/

#include <petscdmda.h>
#include <petscts.h>
#include <petscdraw.h>
#include <petsctime.h>

/* User-defined application context - contains data needed by the
   application-provided call-back routines. */
typedef struct {
  MPI_Comm    comm;      /* communicator */
  DM          da;        /* distributed array data structure */
  Vec         localwork; /* local ghosted work vector */
  Vec         u_local;   /* local ghosted approximate solution vector */
  Vec         solution;  /* global exact solution vector */
  PetscInt    m;         /* total number of grid points in x-direction */
  PetscInt    n;         /* total number of grid points in y-direction */
  PetscReal   h;         /* mesh width h = 1/(m-1) */
  PetscReal   dt;        /* timestep width delta_t */
  PetscReal     ftime;
} AppCtx;

/*
   User-defined routines
*/
extern PetscErrorCode InitialConditions(Vec, AppCtx*, Vec, Vec);
extern PetscErrorCode RHSMatrixHeat(TS, Vec, Mat*, MatStructure*, void*);
extern PetscErrorCode RHSFunctionHeat(TS, PetscReal, Vec, Vec, void*);
extern PetscErrorCode Monitor(TS,PetscInt, PetscReal, Vec, void*);
extern PetscErrorCode ExactSolution(PetscReal, Vec, AppCtx*);

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **argv)
{
  AppCtx         appctx;                 /* user-defined application context */
  TS             ts;                     /* timestepping context */
  Mat            A;                      /* matrix data structure */
  Vec            u;                      /* approximate solution vector */
  Vec            xx, yy;                 /* mesh in x and y direction */
  PetscReal      time_total_max = 1.0;   /* default max total time */
  PetscInt       time_steps_total = 5000;
  PetscInt       time_steps_max = 100;   /* default max timesteps */
  PetscDraw      draw;                   /* drawing context */
  PetscErrorCode ierr;
  PetscInt       steps, m, n;
  PetscMPIInt    size;
  PetscReal      dt, ftime;
  PetscBool      flg;
  TSProblemType  tsproblem = TS_LINEAR;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Initialize program and set problem parameters
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr        = PetscInitialize(&argc,&argv,(char*)0,help);CHKERRQ(ierr);
  appctx.comm = PETSC_COMM_WORLD;

  m        = 60;
  n        = 60;
  ierr     = PetscOptionsGetInt(NULL,"-m",&m,NULL);CHKERRQ(ierr);
  ierr     = PetscOptionsGetInt(NULL,"-n",&n,NULL);CHKERRQ(ierr);
  ierr     = PetscOptionsGetInt(NULL,"-t",&time_steps_max,NULL);
  CHKERRQ(ierr);
  ierr     = PetscOptionsGetInt(NULL,"-tt",&time_steps_total,NULL);
  CHKERRQ(ierr);
  appctx.m = m;
  appctx.n = appctx.m; //n
  n = m;
  appctx.h = 1.0/(m-1.0);

  ierr = MPI_Comm_size(PETSC_COMM_WORLD,&size);CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"Solving a linear TS problem, number of " 
                     "processors = %d\n", size); CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create vector data structures
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* Create distributed array (DMDA) to manage parallel grid and vectors
     and to set up the ghost point communication pattern.  There are MxN
     total grid values spread equally among all the processors. Stencil width 
     is set to one. */
  ierr = DMDACreate2d(PETSC_COMM_WORLD, DMDA_BOUNDARY_NONE, DMDA_BOUNDARY_NONE, 
                      DMDA_STENCIL_STAR, m, n, PETSC_DECIDE, PETSC_DECIDE, 1, 1, 
                      NULL, NULL, &appctx.da); CHKERRQ(ierr);
  /* Extract global and local vectors from DMDA; we use these to store the
     approximate solution.  Then duplicate these for remaining vectors that
     have the same types. */
  ierr = DMCreateGlobalVector(appctx.da,&u);CHKERRQ(ierr);
  ierr = DMCreateLocalVector(appctx.da,&appctx.u_local);CHKERRQ(ierr);
  PetscInt mysize;
  VecGetSize(u,&mysize);
  PetscPrintf(PETSC_COMM_WORLD, "vec size = %d, time_steps_max = %d \n", 
              mysize, time_steps_max);

  /* Create local work vector for use in evaluating right-hand-side function;
     create global work vector for storing exact solution. */
  ierr = VecDuplicate(appctx.u_local, &appctx.localwork); CHKERRQ(ierr);
  ierr = VecDuplicate(u, &appctx.solution); CHKERRQ(ierr);
  ierr = VecDuplicate(u, &xx); CHKERRQ(ierr);
  ierr = VecDuplicate(u, &yy); CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create timestepping solver context
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr = TSCreate(PETSC_COMM_WORLD,&ts);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Set solution vector and initial timestep
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  //dt   = appctx.h*appctx.h/2.0;
  dt = 1.0/time_steps_total;
  appctx.dt = dt;
  ierr = TSSetInitialTimeStep(ts,0.0,dt);CHKERRQ(ierr);
  ierr = TSSetSolution(ts,u);CHKERRQ(ierr);

  ierr = PetscPrintf(PETSC_COMM_WORLD,"dt = %f \t h = %f \n", dt, appctx.h);
  CHKERRQ(ierr);
  double ratio = dt / (2.0*appctx.h*appctx.h);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"dt/(h^2+h^2) = %f \n", ratio);
  CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Create matrix data structure; set matrix evaluation routine.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr = MatCreate(PETSC_COMM_WORLD,&A);CHKERRQ(ierr);
  ierr = MatSetSizes(A,PETSC_DECIDE,PETSC_DECIDE,m*n,m*n);CHKERRQ(ierr);
  ierr = MatSetFromOptions(A);CHKERRQ(ierr);
  ierr = MatMPIAIJSetPreallocation(A,5,NULL,5,NULL);CHKERRQ(ierr);
  ierr = MatSeqAIJSetPreallocation(A,5,NULL);CHKERRQ(ierr);
  ierr = MatSetUp(A);CHKERRQ(ierr);

  /* For linear problems with a time-independent f(u) in the equation
     u_t = f(u), the user provides the discretized right-hand-side
     as a matrix only once, and then sets a null matrix evaluation
     routine. */
  MatStructure A_structure;
  ierr = RHSMatrixHeat(ts, u, &A, &A_structure, &appctx); CHKERRQ(ierr);
  ierr = TSSetRHSFunction(ts, NULL, TSComputeRHSFunctionLinear, &appctx);
  CHKERRQ(ierr);
  ierr = TSSetRHSJacobian(ts, A, A, TSComputeRHSJacobianConstant, &appctx);
  CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Customize timestepping solver:
       - Set the solution method to be the Backward Euler method.
       - Set timestepping duration info
     Then set runtime options, which can override these defaults.
     For example,
          -ts_max_steps <maxsteps> -ts_final_time <maxtime>
     to override the defaults set by TSSetDuration().
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr = TSSetDuration(ts,time_steps_max,time_total_max); CHKERRQ(ierr);
  ierr = TSSetFromOptions(ts);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Solve the problem
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* Evaluate initial conditions */
  ierr = InitialConditions(u, &appctx, xx, yy);CHKERRQ(ierr);
  /* Run the timestepping solver */
  PetscLogDouble t0,t1;
  //PetscTime(&t0);
  t0 = MPI_Wtime();
  ierr = TSSolve(ts,u);CHKERRQ(ierr);
  ierr = TSGetSolveTime(ts,&ftime);CHKERRQ(ierr);
  ierr = TSGetTimeStepNumber(ts,&steps);CHKERRQ(ierr);
  t1 = MPI_Wtime();
  //PetscTime(&t1);
  PetscPrintf(PETSC_COMM_WORLD, "it took this: %g\n",t1-t0);
  TSProblemType type;
  ierr = TSGetProblemType(ts, &type); CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     View timestepping solver info
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr = PetscPrintf(PETSC_COMM_WORLD,"Total timesteps %D, Final time %G\n",
                     steps,ftime);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Output mesh in x and y direction and solution.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
/*  PetscViewer ASCIIviewer, ASCIIviewer_xx, ASCIIviewer_yy;
  PetscViewerASCIIOpen(PETSC_COMM_WORLD, "value.data", &ASCIIviewer);
  PetscViewerASCIIOpen(PETSC_COMM_WORLD, "xx.data", &ASCIIviewer_xx);
  PetscViewerASCIIOpen(PETSC_COMM_WORLD, "yy.data", &ASCIIviewer_yy);
  VecView(u, ASCIIviewer);
  VecView(xx, ASCIIviewer_xx);
  VecView(yy, ASCIIviewer_yy);
  PetscViewerDestroy(&ASCIIviewer);
  PetscViewerDestroy(&ASCIIviewer_xx);
  PetscViewerDestroy(&ASCIIviewer_yy); 
*/

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Free work space. All PETSc objects should be destroyed when they
     are no longer needed.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr = TSDestroy(&ts); CHKERRQ(ierr);
  ierr = MatDestroy(&A); CHKERRQ(ierr);
  ierr = VecDestroy(&u); CHKERRQ(ierr);
  ierr = VecDestroy(&xx); CHKERRQ(ierr);
  ierr = VecDestroy(&appctx.localwork); CHKERRQ(ierr);
  ierr = VecDestroy(&appctx.solution); CHKERRQ(ierr);
  ierr = VecDestroy(&appctx.u_local); CHKERRQ(ierr);
  ierr = DMDestroy(&appctx.da); CHKERRQ(ierr);

  /* Always call PetscFinalize() before exiting a program.  This routine
       - finalizes the PETSc libraries as well as MPI
       - provides summary and diagnostic information if certain runtime
         options are chosen (e.g., -log_summary). */
  ierr = PetscFinalize();
  return 0;
}

/* --------------------------------------------------------------------- */
#undef __FUNCT__
#define __FUNCT__ "InitialConditions"
/* InitialConditions - Computes the solution at the initial time.

   Input Parameter:
   u - uninitialized solution vector (global)
   appctx - user-defined application context

   Output Parameter:
   u - vector with solution at initial time (global) */
PetscErrorCode InitialConditions(Vec u, AppCtx *appctx, Vec xx, Vec yy)
{
  PetscScalar    *u_localptr, h = appctx->h, *xx_localptr, *yy_localptr;
  PetscInt       i, j, k, mybase, myend, m = appctx->m, n = appctx->n;
  PetscErrorCode ierr;

  /* Determine starting point of each processor's range of grid values. */
  ierr = VecGetOwnershipRange(u, &mybase, &myend); CHKERRQ(ierr);

  /* Get a pointer to vector data.
    - For default PETSc vectors, VecGetArray() returns a pointer to
      the data array.  Otherwise, the routine is implementation dependent.
    - You MUST call VecRestoreArray() when you no longer need access to
      the array.
    - Note that the Fortran interface to VecGetArray() differs from the
      C version.  See the users manual for details. */
  ierr = VecGetArray(u, &u_localptr); CHKERRQ(ierr);
  ierr = VecGetArray(xx, &xx_localptr); CHKERRQ(ierr);
  ierr = VecGetArray(yy, &yy_localptr); CHKERRQ(ierr);

  /* We initialize the solution array by simply writing the solution
     directly into the array locations.  Alternatively, we could use
     VecSetValues() or VecSetValuesLocal(). */
  double val;
  for (k = mybase; k < myend; k++) {
    i = k/n; j = k - i*n;
    val = i*h * (i*h-0.5)*(i*h-0.5) * (1-i*h)
         *j*h * (j*h-0.5)*(j*h-0.5) * (1-j*h);
    u_localptr[k-mybase] = val;
    xx_localptr[k-mybase] = i*h;
    yy_localptr[k-mybase] = j*h;
  }

  /* Restore vector */
  ierr = VecRestoreArray(u, &u_localptr); CHKERRQ(ierr);
  ierr = VecRestoreArray(xx, &xx_localptr); CHKERRQ(ierr);
  ierr = VecRestoreArray(yy, &yy_localptr); CHKERRQ(ierr);

  return 0;
}


/* --------------------------------------------------------------------- */
#undef __FUNCT__
#define __FUNCT__ "RHSMatrixHeat"
/*
   RHSMatrixHeat - User-provided routine to compute the right-hand-side
   matrix for the heat equation.

   Input Parameters:
   ts - the TS context
   t - current time
   global_in - global input vector
   dummy - optional user-defined context, as set by TSetRHSJacobian()

   Output Parameters:
   AA - Jacobian matrix
   BB - optionally different preconditioning matrix
   str - flag indicating matrix structure

  Notes:
  RHSMatrixHeat computes entries for the locally owned part of the system.
   - Currently, all PETSc parallel matrix formats are partitioned by
     contiguous chunks of rows across the processors.
   - Each processor needs to insert only elements that it owns
     locally (but any non-local elements will be sent to the
     appropriate processor during matrix assembly).
   - Always specify global row and columns of matrix entries when
     using MatSetValues(); we could alternatively use MatSetValuesLocal().
   - Here, we set all entries for a particular row at once.
   - Note that MatSetValues() uses 0-based row and column numbers
     in Fortran as well as in C.
*/
PetscErrorCode RHSMatrixHeat(TS ts, Vec X, Mat *AA, 
                             MatStructure *str, void *ctx)
{
  Mat            A       = *AA;          /* Jacobian matrix */
  AppCtx         *appctx = (AppCtx*)ctx; /* user-defined application context */
  PetscErrorCode ierr;
  PetscInt       i, Istart, Iend, Ii, Ij, j, J, m, n;

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Compute entries for the locally owned part of the matrix
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  ierr = MatGetOwnershipRange(A, &Istart, &Iend); CHKERRQ(ierr);

  /// u^{n+1} = [-4u^n_{i,j} + u^n_{i-1,j} + u^n_{i-1,j} + u^n_{i,j-1} + 
  ///            u^n_{i,j+1}]*(delta_t/h^2) + u^n_{i,j} [+ f_{i,j}*delta_t]
  m = appctx->m; n = appctx->n;
  PetscScalar u, w;
  u = 1.0/(appctx->h*appctx->h);
  w = -4.0/(appctx->h*appctx->h);
  /* Set matrix elements for the 2-D, five-point stencil in parallel.
      - Each processor needs to insert only elements that it owns
        locally (but any non-local elements will be sent to the
        appropriate processor during matrix assembly).
      - Always specify global rows and columns of matrix entries. */
  for (Ii=Istart; Ii<Iend; Ii++) {
    i = Ii/n; j = Ii - i*n;
    if (i>0)   {J = Ii - n; ierr = MatSetValues(A,1,&Ii,1,&J,&u,INSERT_VALUES);
                CHKERRQ(ierr);}
    if (i<m-1) {J = Ii + n; ierr = MatSetValues(A,1,&Ii,1,&J,&u,INSERT_VALUES);
                CHKERRQ(ierr);}
    if (j>0)   {J = Ii - 1; ierr = MatSetValues(A,1,&Ii,1,&J,&u,INSERT_VALUES);
                CHKERRQ(ierr);}
    if (j<n-1) {J = Ii + 1; ierr = MatSetValues(A,1,&Ii,1,&J,&u,INSERT_VALUES);
                CHKERRQ(ierr);}
    ierr = MatSetValues(A,1,&Ii,1,&Ii,&w,INSERT_VALUES);CHKERRQ(ierr);
  }
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
     Complete the matrix assembly process and set some options
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* Assemble matrix, using the 2-step process:
       MatAssemblyBegin(), MatAssemblyEnd()
     Computations can be done while messages are in transition
     by placing code between these two statements. */
  ierr = MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);

  /* Set flag to indicate that the Jacobian matrix retains an identical
     nonzero structure throughout all timestepping iterations (although the
     values of the entries change). Thus, we can save some work in setting
     up the preconditioner (e.g., no need to redo symbolic factorization for
     ILU/ICC preconditioners).
      - If the nonzero structure of the matrix is different during
        successive linear solves, then the flag DIFFERENT_NONZERO_PATTERN
        must be used instead.  If you are unsure whether the matrix
        structure has changed or not, use the flag DIFFERENT_NONZERO_PATTERN.
      - Caution:  If you specify SAME_NONZERO_PATTERN, PETSc
        believes your assertion and does not check the structure
        of the matrix.  If you erroneously claim that the structure
        is the same when it actually is not, the new preconditioner
        will not function correctly.  Thus, use this optimization
        feature with caution! */
  *str = SAME_NONZERO_PATTERN;

  /* Set and option to indicate that we will never add a new nonzero location
     to the matrix. If we do, it will generate an error. */
  ierr = MatSetOption(A,MAT_NEW_NONZERO_LOCATION_ERR,PETSC_TRUE); CHKERRQ(ierr);

  return 0;
}
