/** scafes_vs_petsc
 *  Copyright (c) 2013, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#include <iostream>
#include "ScaFES.hpp"
template<typename CT, std::size_t DIM> // Source f.
inline void funcF(CT& fx, ScaFES::Ntuple<CT,DIM> const& x, CT const& t) {
    fx = 0.0;
}
template<typename CT, std::size_t DIM> // Boundary condition g.
inline void funcG(CT& fx, ScaFES::Ntuple<CT,DIM> const& x, CT const& t) {
    fx = 0.0;
}
template<typename CT, std::size_t DIM> // Initial condition \tilde{u}.
inline void funcUt(CT& fx, ScaFES::Ntuple<CT,DIM> const& x, CT const& t) {
    fx = 1.0;
    for (std::size_t pp = 0; pp < DIM; ++pp) {
        fx *= (x[pp] * (x[pp] - 0.5) * (x[pp] - 0.5) * (1.0 - x[pp]));
    }
}
template<typename CT, std::size_t DIM> // Own problem class.
class HeatEqnFDM : public ScaFES::Problem<HeatEqnFDM<CT,DIM>, CT, DIM> {
   public:
      HeatEqnFDM(ScaFES::Parameters const& cl,
                 ScaFES::GridGlobal<DIM> const& gg)
         : ScaFES::Problem<HeatEqnFDM<CT,DIM>, CT, DIM>(cl, gg) {
         this->addDataField("F", 0, funcF<CT,DIM>, true);
         this->addDataField("G", 0, funcG<CT,DIM>, true);
         this->addDataField("U", 1, funcUt<CT,DIM>, false);
      }
      template<typename TT> // Method must be implemented!
      void updateInner(std::vector<ScaFES::DataField<TT,DIM>>& vNew,
                       std::vector<ScaFES::DataField<TT,DIM>> const& vOld,
                       ScaFES::Ntuple<int,DIM> const& idxNode,
                       int const& timestep) {
         vNew[0](idxNode) = vOld[0](idxNode)
                            - this->tau() * this->knownDf(0, idxNode);
         for (std::size_t pp = 0; pp < DIM; ++pp) {
            vNew[0](idxNode) -= this->tau() * (
                 2.0 * vOld[0](idxNode)
                     - vOld[0](this->connect(idxNode, 2*pp))
                     - vOld[0](this->connect(idxNode, 2*pp+1)) )
                     / (this->gridsize(pp) * this->gridsize(pp));
         }
      }
      template<typename TT> // Method must be implemented!
      void updateBorder(std::vector<ScaFES::DataField<TT,DIM>>& vNew,
                        std::vector<ScaFES::DataField<TT,DIM>>const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep) {
          vNew[0](idxNode) = this->knownDf(1, idxNode);
      }
      template<typename TT> // Method must be implemented!
      void initialise(std::vector<ScaFES::DataField<TT,DIM>>& vNew,
                      std::vector<ScaFES::DataField<TT,DIM>>const& vOld,
                      ScaFES::Ntuple<int,DIM> const& idxNode,
                      int const& timestep) {
      }
      template<typename TT> // Method must be implemented!
      void initialise(std::vector<ScaFES::DataField<TT,DIM>>& vNew,
                        std::vector<TT>const& vOld,
                        ScaFES::Ntuple<int,DIM> const& idxNode,
                        int const& timestep) {
      }
};
int main(int argc, char *argv[]) {      // Main program.
   ScaFES::Parameters pp(argc, argv);   // Read in command line options.
   ScaFES::GridGlobal<2> gg(pp);        // Create grid partitions.
   HeatEqnFDM<double,2> prblm(pp, gg);  // Create 2D heat eqn. problem.
   double sumHsquared = 0.0;
   for (std::size_t pp = 0; pp < 2; ++pp) {
       sumHsquared += prblm.gridsize(pp) * prblm.gridsize(pp);
   }
   std::cout << " Stability assumption: " 
             << prblm.tau() / sumHsquared << std::endl;
   if (0.5 < prblm.tau() / sumHsquared && 0 == prblm.myRank()) {
       std::cout << "\n Stability assumption is not fulfilled." << std::endl;
   }
   prblm.iterate();                     // Iterate over all time steps.
   return 0;
}
